import struct
from PIL import Image
import os
import glob
import sys

# fn - absolute filename
# an - absolute path with name (no extension)
# ad - absolute path to directory
# rn - relative filename

def binary(short):
	st = ""
	for ibit in range(15,-1,-1):
		if (1 << ibit) & short == 0:
			st = st + "0"
		else:
			st = st + "1"
	return st
	  
class Pic(object):
	"Holds a pil image and a filename, and can save the image."
	def __init__(self, image, fn):
		self.image = image
		self.fn = fn
		self.x = 0
		self.y = 0
		(self.w, self.h) = image.size
		
	def Save(self):
		print "Saving " + os.path.basename(self.fn)
		self.image.save(self.fn, transparency=0)

	def PutPixel(self, col):
		"Put a pixel in the image, autoincrementing x and y."
		if not self.FMore():
			self.Save()
			raise Exception("putpixel past end of picp!")
			
		self.image.putpixel((self.x, self.y), col)
		self.x = self.x + 1
		if self.x == self.w:
			self.x = 0
			self.y = self.y + 1
			
	def FMore(self):
		"Returns true if there are any more pixels to put."
		if self.y == self.h:
			return False
		return True

class Pal(object):
	def __init__(self, fnPngPal):
		self.image = Image.open(fnPngPal)
	
	def Palette(self):
		"Generate a palette for pil (list of 768 integers) based on a colour grid."
		(w, h) = self.image.size
		wBox = w / 16
		hBox = h / 16
		palette = []
		# {x,y}p - "palette-based" xy value (0-15)
		for xp in range(16):
			for yp in range(16):
				palette.extend(self.image.getpixel((xp * wBox, yp * hBox)))
		return palette

class KnpExtractor(object):
	def __init__(self, an, pal):
		self.pal = pal
		self.an = an
		
	def FnImg(self):
		return self.an + ".img"
	
	def AdOut(self):
		return self.an
		
	def FnPic(self, rn):
		return self.AdOut() + "/" + rn + ".png"
					
	def Extract(self):
		if not os.path.exists(self.AdOut()):
			os.mkdir(self.AdOut())
		with open(self.FnImg(), "rb") as f:
			rfFile = f.read()
		(cimgHdr,) = struct.unpack_from("<I", rfFile);
		foImgHdr = 4
		for iimgHdr in range(cimgHdr):
			(foImg, szImg) = struct.unpack_from("<II", rfFile, foImgHdr)
			if foImg > 0:
				self.PicFromRfImg(iimgHdr, rfFile[foImg:foImg + szImg]).Save()
			foImgHdr = foImgHdr + 8
		
	def PicFromRfImg(self, iimg, rfImg):
		(width, height) = struct.unpack_from("<HH", rfImg, 10)
		(cmpf, hotX, hotY) = struct.unpack_from("<BHH", rfImg, 15)
		rn = str(iimg) + "_" + str(hotX) + "_" + str(hotY)
		if (width % 2) == 1:
			width = width + 1
		image = Image.new("P", (width, height), 0)
		image.putpalette(self.pal.Palette())
		pic = Pic(image, self.FnPic(rn))
		rfImdat = rfImg[24:-1] # remove trailing zero
		
		if cmpf == 0:
			self.PutImageRaw(pic, rfImdat)
		elif cmpf == 1:
			self.PutImageRle(pic, rfImdat)
		else:
			raise Exception("Invalid compression format " + cmpf)
		
		return pic
		
	def PutImageRaw(self, pic, rfImdat):
		pic.image.putdata(rfImdat)
		
	def PutImageRle(self, pic, rfImdat):
		irfImdat = 0
		while pic.FMore():
			ctl = ord(rfImdat[irfImdat])
			irfImdat = irfImdat + 1
			if (ctl & 0x80) == 0: # rle
				cpixSame = ctl
				col = ord(rfImdat[irfImdat])
				for ipixSame in range(cpixSame):
					pic.PutPixel(col)
				irfImdat = irfImdat + 1
			else:
				cpixUnique = ctl & 0x7f
				for ipixUnique in range(cpixUnique):
					pic.PutPixel(ord(rfImdat[irfImdat]))
					irfImdat = irfImdat + 1

def main(fnExe, rgarg):
	print "KNPExtract v0.1"
	print "By Jeremy Penner"
	print "This code is public domain"
	print
	print "Usage:"
	print "KNPExtract [directory to scan for .img files | .img file] ..."
	
	pal = Pal(os.path.dirname(fnExe) + "/knp_pal.png")
	if (len(rgarg) == 0):
		ScanAd(os.getcwd(), pal)
	else:
		for arg in rgarg:
			fnImg = AnFromFn(arg) + ".img"
			if os.path.exists(fnImg):
				Extract(fnImg, pal)
			elif os.path.isdir(arg):
				ScanAd(arg, pal)

def ScanAd(ad, pal):
	for fnImg in glob.glob(ad + "/*.img"):
		Extract(fnImg, pal)

def AnFromFn(fn):
	if fn.lower().endswith(".img"):
		return fn[:-4]
	return fn
	
def Extract(fn, pal):
	print
	print "Extracting from " + os.path.basename(fn) + "..."
	try:
		KnpExtractor(AnFromFn(fn), pal).Extract()
	except e:
		print "Error: " + e
		
if __name__ == "__main__":
	main(os.path.abspath(sys.argv[0]), sys.argv[1:])